"""Classes for defining optimization problem objects for Simulated Annealing (SA)."""

# Author: Genevieve Hayes
# License: BSD 3-clause

from .arithmetic_decay import ArithmeticDecay
from .custom_decay import CustomDecay
from .exponential_decay import ExponentialDecay
from .geometric_decay import GeometricDecay
