"""Classes for synthetic data generation and visualization."""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

from .synthetic_data import SyntheticDataGenerator, plot_synthetic_dataset
