"""Functions for defining decorators to be used throughout the code base."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from .short_name_decorator import short_name, get_short_name
