"""Classes for generating optimization problem objects."""

# Authors: Genevieve Hayes (modified by Kyle Nakamura)
# License: BSD 3-clause

from .continuous_peaks_generator import ContinuousPeaksGenerator
from .flip_flop_generator import FlipFlopGenerator
from .four_peaks_generator import FourPeaksGenerator
from .knapsack_generator import KnapsackGenerator
from .max_k_color_generator import MaxKColorGenerator
from .one_max_generator import OneMaxGenerator
from .queens_generator import QueensGenerator
from .six_peaks_generator import SixPeaksGenerator
from .tsp_generator import TSPGenerator
