"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from .continuous_opt import ContinuousOpt
from .discrete_opt import DiscreteOpt
from .flip_flop_opt import FlipFlopOpt
from .knapsack_opt import KnapsackOpt
from .max_k_color_opt import MaxKColorOpt
from .queens_opt import QueensOpt
from .tsp_opt import TSPOpt
